#!/usr/bin/env python3

# 
# CCDD: Constrained Conditional Denoising Diffusion
# Copyright (C) 2025  James E. Fowler
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
# MA 02139, USA.
# 


import os
import sys
from Data.Pavia import make_data as pavia
from Data.Chikusei import make_data as chikusei
from Data.LosAngeles import make_data as losangeles


def MakeData():
    print("Processing Pavia University")
    pavia.MakeData()
    print("Processing Chikusei")
    chikusei.MakeData()
    print("Processing Los Angeles")
    losangeles.MakeData()


if __name__ == "__main__":
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, dirname)
    MakeData()
